@@ -6,6 +6,8 @@
 import collections
 import enum
 import fnmatch
+import functools
+import operator
 import os
 import re
 import warnings
@@ -27,7 +29,7 @@ INT_DEFAULTS = dict.fromkeys((
     'min_dh_size', 'min_dsa_size', 'min_rsa_size',
     'sha1_in_certs',
     'ssh_certs',
-), 0)
+), 0) | {'min_ec_size': 256}
 
 
 # For enum values, first value works as default,
@@ -35,6 +37,7 @@ INT_DEFAULTS = dict.fromkeys((
 ENUMS = {
     'etm': ('ANY', 'DISABLE_ETM', 'DISABLE_NON_ETM'),
     '__ems': ('DEFAULT', 'ENFORCE', 'RELAX'),  # FIPS/NO-ENFORCE-EMS
+    '__rsaes_pkcs1_v1_5_encrypt_decrypt': ('DISALLOW', 'ALLOW'),  # LEGACY
 }
 
 
@@ -43,7 +46,11 @@ ENUMS = {
 SCOPE_ANY = '*'
 
 ALL_SCOPES = (  # defined explicitly to catch typos / globbing nothing
-    'tls', 'ssl', 'openssl', 'nss', 'gnutls', 'java-tls',
+    'tls', 'ssl',
+    'openssl', 'gnutls', 'java-tls',
+    'nss', 'nss-tls',
+    'pkcs12', 'pkcs12-import', 'nss-pkcs12', 'nss-pkcs12-import',
+    'smime', 'smime-import', 'nss-smime', 'nss-smime-import',
     'ssh', 'openssh', 'openssh-server', 'openssh-client', 'libssh',
     'ipsec', 'ike', 'libreswan',
     'kerberos', 'krb5',
@@ -51,18 +58,27 @@ ALL_SCOPES = (  # defined explicitly to catch typos / globbing nothing
     'sequoia', 'rpm', 'rpm-sequoia',
 )
 DUMPABLE_SCOPES = {  # TODO: fix duplication, backends specify same things
-    'bind': {'bind', 'dnssec'},
-    'gnutls': {'gnutls', 'tls', 'ssl'},
-    'java-tls': {'java-tls', 'tls', 'ssl'},
-    'krb5': {'krb5', 'kerberos'},
-    'libreswan': {'ipsec', 'ike', 'libreswan'},
-    'libssh': {'libssh', 'openssh', 'ssh'},
-    'nss': {'nss', 'tls', 'ssl'},
-    'openssh-client': {'openssh-client', 'openssh', 'ssh'},
-    'openssh-server': {'openssh-server', 'openssh', 'ssh'},
-    'openssl': {'openssl', 'tls', 'ssl'},
-    'sequoia': {'sequoia'},
-    'rpm': {'rpm', 'rpm-sequoia'},
+    # scope we dump under: (relative parent scope, set of scopes applied)
+    'bind': (None, {'bind', 'dnssec'}),
+    'gnutls': (None, {'gnutls', 'tls', 'ssl'}),
+    'java-tls': (None, {'java-tls', 'tls', 'ssl'}),
+    'krb5': (None, {'krb5', 'kerberos'}),
+    'libreswan': (None, {'ipsec', 'ike', 'libreswan'}),
+    'libssh': (None, {'libssh', 'ssh'}),
+    'nss': (None, {'nss'}),
+    'nss-tls': ('nss', {'nss', 'nss-tls', 'tls', 'ssl'}),
+    'nss-pkcs12': ('nss', {'nss', 'pkcs12', 'nss-pkcs12'}),
+    'nss-pkcs12-import': ('nss-pkcs12', {'nss', 'pkcs12', 'pkcs12-import',
+                                         'nss-pkcs12', 'nss-pkcs12-import'}),
+    'nss-smime': ('nss', {'nss', 'smime', 'nss-smime'}),
+    'nss-smime-import': ('nss-smime', {'nss', 'smime', 'smime-import',
+                                       'nss-smime', 'nss-smime-import'}),
+    'openssh': (None, {'openssh', 'ssh'}),
+    'openssh-client': ('openssh', {'openssh-client', 'openssh', 'ssh'}),
+    'openssh-server': ('openssh', {'openssh-server', 'openssh', 'ssh'}),
+    'openssl': (None, {'openssl', 'tls', 'ssl'}),
+    'sequoia': (None, {'sequoia'}),
+    'rpm': (None, {'rpm', 'rpm-sequoia'}),
 }
 
 
@@ -243,11 +259,12 @@ def syntax_check_line(line, warn=False):
 
 
 class PolicySyntaxDeprecationWarning(FutureWarning, validation.PolicyWarning):
-    def __init__(self, deprecated, replacement):
+    def __init__(self, deprecated, replacement, what='option', onetoone=False):
         replacement = replacement.replace('\n', ' and ')
-        msg = f'option {deprecated} is deprecated'
-        msg += f', please rewrite your rules using {replacement}; '
-        msg += 'be advised that it is not always a 1-1 replacement'
+        msg = f'{what} {deprecated} is deprecated'
+        msg += f', please rewrite your rules using {replacement}'
+        if not onetoone:
+            msg += '; be advised that it is not always a 1-1 replacement'
         super().__init__(msg)
 
 
@@ -266,7 +283,13 @@ def preprocess_text(text):
     ...     preprocess_text('min_tls_version=TLS1.3')
     'protocol@TLS = -SSL2.0 -SSL3.0 -TLS1.0 -TLS1.1 -TLS1.2'
     """
-    text = re.sub(r'#.*', '', text)
+    magic_comments = {
+        '# %suppress_experimental_value_warnings=true',
+        '# %suppress_experimental_value_warnings=false',
+    }
+    text = re.sub(r'#.*',
+                  lambda s: s.group(0) if s.group(0) in magic_comments else '',
+                  text)
     text = text.replace('=', ' = ')
     text = '\n'.join(l.strip() for l in text.split('\n'))
     text = text.replace('\\\n', '')
@@ -310,11 +333,25 @@ def preprocess_text(text):
         regex = r'\b' + fr + r'\b'
         matches = {}
         for match in re.finditer(regex, text):
-            matches[match.group(0)] = re.sub(regex, to, text)
+            matches[match.group(0)] = re.sub(regex, to, match.group(0))
         for match_fr, match_to in matches.items():
             warnings.warn(PolicySyntaxDeprecationWarning(match_fr, match_to))
         text = re.sub(regex, to, text)
 
+    VALUE_REPLACEMENTS = {
+        'X25519-MLKEM768': 'MLKEM768-X25519',  # RHEL-99813
+    }
+    for fr, to in VALUE_REPLACEMENTS.items():
+        regex = r'\b' + fr + r'\b'
+        matches = {}
+        for match in re.finditer(regex, text):
+            matches[match.group(0)] = re.sub(regex, to, match.group(0))
+        for match_fr, match_to in matches.items():
+            warnings.warn(PolicySyntaxDeprecationWarning(match_fr, match_to,
+                                                         what='value',
+                                                         onetoone=True))
+        text = re.sub(regex, to, text)
+
     dtls_versions = list(alg_lists.DTLS_PROTOCOLS[::-1])
     while dtls_versions:
         neg = " ".join("-" + v for v in dtls_versions[:-1])
@@ -436,6 +473,24 @@ class UnscopedCryptoPolicy:
     def scoped(self, scopes=None):
         return ScopedPolicy(self._directives, scopes or {})
 
+    @staticmethod
+    def _for_in_lines_expwarn(lines, func):
+        warn_on_experimental = True  # reset between invocations
+        ExpValWarning = validation.alg_lists.ExperimentalValueWarning
+        r = []
+        for l in lines:
+            if l == '# %suppress_experimental_value_warnings = true':
+                warn_on_experimental = False
+            elif l == '# %suppress_experimental_value_warnings = false':
+                warn_on_experimental = True
+            elif warn_on_experimental:
+                r.append(func(l))
+            else:
+                with warnings.catch_warnings(append=True):
+                    warnings.filterwarnings('ignore', category=ExpValWarning)
+                    r.append(func(l))
+        return r
+
     def read_policy_file(self, name, subpolicy=False):
         pdir = self.policydir or 'policies'
         if subpolicy:
@@ -452,11 +507,16 @@ class UnscopedCryptoPolicy:
             text = f.read()
         text = preprocess_text(text)
         lines = text.split('\n')
-        for l in lines:  # display several warnings at once
-            syntax_check_line(l, warn=True)
-        for l in lines:  # crash
-            syntax_check_line(l)
-        return [x for l in lines for x in parse_line(l)]
+        # display several warnings at once, if there are any
+        self._for_in_lines_expwarn(lines,
+                                   lambda l: syntax_check_line(l, warn=True))
+        # crash, if there are any
+        self._for_in_lines_expwarn(lines, syntax_check_line)
+
+        # return [x for l in lines for x in parse_line(l)],  # noqa: ERA001
+        # but with configurable experimental value warnings uppression
+        r = self._for_in_lines_expwarn(lines, parse_line)
+        return functools.reduce(operator.iadd, r)
 
     def __str__(self):
         def fmt(key, value):
@@ -476,13 +536,21 @@ class UnscopedCryptoPolicy:
         for prop_name, value in generic_all.items():
             s += fmt(prop_name, value)
         anything_scope_specific = False
-        for scope_name, scope_set in DUMPABLE_SCOPES.items():
+        for scope_name, (parent_scope, scope_set) in DUMPABLE_SCOPES.items():
             specific_scoped = self.scoped(scopes=scope_set)
             specific_all = {**specific_scoped.enabled,
                             **specific_scoped.integers,
                             **specific_scoped.enums}
+            if parent_scope is not None:
+                _, parent_scope_set = DUMPABLE_SCOPES[parent_scope]
+                parent_scoped = self.scoped(scopes=parent_scope_set)
+                parent_all = {**parent_scoped.enabled,
+                              **parent_scoped.integers,
+                              **parent_scoped.enums}
+            else:
+                parent_scoped, parent_all = generic_scoped, generic_all
             for prop_name, value in specific_all.items():
-                if value != generic_all[prop_name]:
+                if value != parent_all[prop_name]:
                     if not anything_scope_specific:
                         s += ('# Scope-specific properties '
                               'derived for select backends:\n')
